// ESFrameModelSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;
import swarm.defobj.SymbolImpl;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.SwarmImpl;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.space.Grid2dImpl;
import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.collections.ArrayImpl;
import swarm.collections.ListShufflerImpl;

import java.io.*;
import java.util.*;

/**
 * The Model contains the Units
 *
 * @author Pietro Terna, with contributions of Marco Remondino, Dario Landini, 
 * Nicola Ormezzano and Paolo Pelligra
 */
public class ESFrameModelSwarm extends SwarmImpl
{
    // simulation parameters

    /** the total number of operating Units we are using*/
    public int potentialUnitTypes;
    /** the probability of initial generation of a unit, listed into the 
     *  unit directory */
    public double unitGenerationInitialP;
    /** repetition in unit creation 
     *  if potentialUnitNumberPerType > 1 each type of unit (as in each row of 
     *  unitData?/unitBasicData.txt) can be be created in multiple copies, 
     *  always applying unitGenerationInitialP; unit number are 
     *  unit*100+repetition factor (if  potentialUnitNumberPerType > 99 we have 
     *  nonsense numbers)*/
    public int potentialUnitNumberPerType;
    /* the max number of steps to be done to complete an order */
    /** probability of unit new Unit generation while the simulation is running */
    public double newUnitGenerationP;

    public int maxStepNumber;
    /** the max length of a step in a recipe (in units of time) */
    public int maxStepLength;
    /** the max number of ticks that a product can spent in a unit */
    public int maxTicksAsUnsent;
    /** the total production time used by done orders */
    public float totalProductionTime;
    /** the total length of the recipes of the done orders */
    public float totalRecipeLength;
    /** the X size  of the unit space */
    public int unitSpaceXSize;
    /** the Y size  of the unit space */
    public int unitSpaceYSize;
    /** the unit space (a metaphorical one) */
    public Grid2dImpl unitSpace;
    /** the unit activity space, always metaphoric */
    public UnitActivitySpace unitActivitySpace;
    /** the min starting visibility */
    public double minStartingVisibility;
    /** the max starting visibility */
    public double maxStartingVisibility;
    /** the min level of inter-visibility; below this level two units cannot
     *  exchange products */
    public double interVisibilityMinLevel;
    /** increasing visibility in each step 1 of each unit */
    public double increasingVisibilityStep;
    /** max inactivity value */
    public int maxInactivity;
    /** max quantity of unsent products */
    public int maxUnsentProducts;
    /** the color of the unit */
    public byte unitColor = (byte) 1;
    /** potential production from the beginning of the run */
    public float potentialProduction;
    /** global production from the beginning of the run */
    public float globalProduction;
    /** concluded orders form the beginning of the simulation */
    public float concludedOrders;
    /** create the display of the numberOfUnits */
    public boolean createNumberOfUnitsGraph;
    /** create the display of the production */
    public boolean createProductionGraph;
    /** create the display of the ratios */
    public boolean createRatioGraph;

    /** the stratum of this model and the stratumNumber, i.e. the number of models,
     *  numbered from 0 to stratumNumber-1*/
    public int stratum, stratumNumber;

    /** the order generator */
    public OrderGenerator orderGenerator;

    /** use the order distiller instead of the order generator */
    public boolean useOrderDistiller;

    /** the order distiller, using archives of recipes and of order sequences
     *  to generate the simulated events */
    public OrderDistiller orderDistiller;

    /** 
     * ActionGroup for holding an ordered sequence of action 
     * (the numbers of the model action names are the same used in the full
     * jES model, so some are lacking in this version)
    */
    public ActionGroup modelActions2, modelActions2b, 
	modelActions2generator, modelActions2distiller;

    /** the Schedule operating in the Model*/
    public Schedule modelSchedule;

    /** the input class for unit parameters*/
    public UnitParameters unitParameters;

    /** the list of the operating units */
    public ListImpl unitList;
    /** its iterator */
    public ListIndex unitListIndex;

    /** a shuffler for the list */
    public ListShufflerImpl myShuffler;

    /** manage the lists in each cycle to have a unique assignment in each cycle to
     *  each unit in the AssigningTool; the choice is related to the units belonging to
     *  this stratum */
    public boolean uniqueAssignmentInEachCycle;

    /** shuffle the lists at each assignment in the AssigningTool (this is done beside
     *  the ordinary shuffle operations in each model schedule); the choice is related 
     *  to the units belonging to this stratum */
    public boolean shuffleListsAtEachAssignment;

    /** stop if zero units in this model */
    public boolean stopIfZeroUnits;

    /** the list of the pro forma units (all the "would be" units) */
    public ListImpl proFormaUnitList;

    /** the tool to assign order to units */
    AssigningTool assigningTool;

    /** the array of memory matrixes to be used by computational objects */
    public ArrayImpl matrixArray;

    /** a memory matrix */
    public MemoryMatrix aMatrix;

    /** the list of computationalAssembler units linked to normal unit; 
     *  comp. ass. units have in charge the task of making the computations */
    public ListImpl computationalAssemblerList;

    /** printMatrixes */
    public boolean printMatrixes;

    /** number of memory matrixes */
    public int totalMemoryMatrixNumber;


    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameModelSwarm$1$ESFrameModelProbeMap)
     */
    Object modelProbe;

    // the constructor of the model
    public ESFrameModelSwarm (Zone aZone) {
	super (aZone);

	// Now fill in various simulation parameters with default values.
	useOrderDistiller=false;
	potentialUnitTypes = 10;
	unitGenerationInitialP=0.9;
        potentialUnitNumberPerType=1;
	newUnitGenerationP=0.0;
	maxStepNumber=30;
	maxStepLength=1;
	maxTicksAsUnsent=0;
	unitSpaceXSize=100;
	unitSpaceYSize=100;
	minStartingVisibility=0.0;
	maxStartingVisibility=100.0;
	interVisibilityMinLevel=0.0;
	increasingVisibilityStep=0.0;
	maxInactivity=10;
        maxUnsentProducts=10;
	totalMemoryMatrixNumber=0;
	globalProduction=0;
	potentialProduction=0;
	concludedOrders=0;
	createNumberOfUnitsGraph=true;
	createProductionGraph=true;
	createRatioGraph=true;
	shuffleListsAtEachAssignment=false;
	uniqueAssignmentInEachCycle=false;
	stopIfZeroUnits=true;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameModelSwarm$1$ESFrameModelProbeMap.class
	class ESFrameModelProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameModelSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }
	    public ESFrameModelProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("potentialUnitTypes");
		addVar ("unitGenerationInitialP");
		addVar ("potentialUnitNumberPerType");
		addVar ("newUnitGenerationP");
		addVar ("maxStepNumber");
		addVar ("maxStepLength");
		addVar ("maxTicksAsUnsent");
		addVar ("unitSpaceXSize");
		addVar ("unitSpaceYSize");
		addVar ("minStartingVisibility");
		addVar ("maxStartingVisibility");
		addVar ("interVisibilityMinLevel");
		addVar ("increasingVisibilityStep");
		addVar ("maxInactivity");
		addVar ("maxUnsentProducts");
		addVar ("totalMemoryMatrixNumber");
		addVar ("createNumberOfUnitsGraph");
		addVar ("createProductionGraph");
		addVar ("createRatioGraph");
		addVar ("useOrderDistiller");
		addVar ("shuffleListsAtEachAssignment");
		addVar ("uniqueAssignmentInEachCycle");
		addVar ("stopIfZeroUnits");
	    }
	    
	}

	// Install our custom probeMap class directly into the
	// probeLibrary
	modelProbe=new ESFrameModelProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameModelProbeMap) modelProbe, getClass ());
    }
    /** Build model objects. */
    public Object buildObjects ()
    {
	int i, rep;

	// allow our parent class to build anything
	super.buildObjects();

	// check visibility
	if(maxStartingVisibility<minStartingVisibility)
	    {
		System.out.println(
				   "maxStartingVisibility < minStartingVisibility");
		MyExit.exit(1);
	    }
			

        // create the instance to read acounting parameters
        unitParameters = new 
	    UnitParameters(getZone(), potentialUnitTypes, stratum);
	unitParameters.readUnitData();

	// the list of the production units
	unitList = new ListImpl (getZone());
	// its iterator
	unitListIndex = unitList.listBegin(getZone());

	// the shuffler
	myShuffler= new ListShufflerImpl(getZone());

	// the list of the pro forma units
	proFormaUnitList = new ListImpl (getZone());

	// create the unit activity space (which is automatically filled with 0 */
	unitActivitySpace = new UnitActivitySpace(Globals.env.globalZone, 
					      unitSpaceXSize, unitSpaceYSize);

	// create a 2-D grid that will be used to keep track of
	// each unit metaphorical or real position
	unitSpace = new Grid2dImpl(Globals.env.globalZone, 
				   unitSpaceXSize, unitSpaceYSize);
	unitSpace.fastFillWithObject(null);

	// creating arrays and vectors in the information repository used to distribute 
	// information among models (done once in a run)
	if(stratum==0) InformationRepository.createArraysAndVectors(getZone(),
								  stratumNumber);
	// setting information
	InformationRepository.setInformationFromModel(stratum, unitList, unitListIndex,
						      unitActivitySpace, interVisibilityMinLevel,
						      uniqueAssignmentInEachCycle, 
						      shuffleListsAtEachAssignment);

	// the instance of assigningTool
	assigningTool = new AssigningTool (getZone(), stratumNumber, stratum
					   /* unitList, unitListIndex,
					   unitActivitySpace, interVisibilityMinLevel*/
					   );

	// the list of the computational assemblers linked to the units 
	// that will be generated 
	computationalAssemblerList = new ListImpl (getZone());

	// the orderGenerator or the orderDistiller (with pro forma unit list to build 
	// the full dictionary
	if(! useOrderDistiller)
	    orderGenerator = new OrderGenerator
		(getZone(), maxStepNumber, maxStepLength, proFormaUnitList,
		 this, assigningTool);
	else
	    orderDistiller = new OrderDistiller
		(getZone(), maxStepNumber, maxStepLength, proFormaUnitList,
		 this, assigningTool, stratum);

	// create the pro forma units (used to build the activity dictionary (containing 
	// the steps to be included in the recipes) in OrderGenerator
	if (potentialUnitTypes==0)
	    {
		System.out.println("General error: potentialUnitTypes cannot be set to zero");
		MyExit.exit(1);
	    }

	for (i = 1; i <= potentialUnitTypes; i++)
	    {
		Unit aProFormaUnit;
		
		aProFormaUnit = new Unit  (getZone(), this, unitActivitySpace, 
					   assigningTool, (int) 0, (int) 0, (byte) 0, 
					   (int) 0, (int) 0, null, null, maxTicksAsUnsent);
		unitParameters.setParametersTo(i, 0, aProFormaUnit);
		
		// add the unit to the list
		proFormaUnitList.addLast (aProFormaUnit);
	    }



	// create the units
	// the following line is commented to allow the creation of empty strata
	// if (potentialUnitNumberPerType == 0) potentialUnitNumberPerType=1;
	for (i = 1; i <= potentialUnitTypes; i++)
	    for (rep=1; rep <= potentialUnitNumberPerType; rep++)
		if(unitGenerationInitialP >= 
		   Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
		    generateUnit(i, rep);

  	/* the following block has been commented to allow the creation of empty strata
	if (unitList.getCount()==0)
	    {
		System.out.println("General error, no unit generated");
		MyExit.exit(1);
	    }
	*/

	// creating the dictionary in orderGenerator or in OrderDistiller
	if(! useOrderDistiller)	orderGenerator.setDictionary();
	else                    orderDistiller.setDictionary();


	// the actual memoryMatrixNumber
	if(stratum>0)totalMemoryMatrixNumber=StartESFrame.eSFrameObserverSwarm.
		       getTotalMemoryMatrixNumber();

	if(totalMemoryMatrixNumber>10)
	    {
		System.out.println("The total memory matrix number cannot be >10.\n"+
				   "If you need to create more matrixes, modify the\n"+
				   "code in ESFrameModelSwarm.java and in "+
				   "ComputationalAssemblerBasic.java\n"+
				   "(both the matrix definitions and the 1100 step).");
		MyExit.exit(1);
	    }

	// creating memory matrixes
	if(totalMemoryMatrixNumber>0){
	matrixArray = new ArrayImpl(getZone(), totalMemoryMatrixNumber);

	if(stratum==0)
	    {
		// reading memory matrix dimension from unitData0/memoryMatrixes.txt
		// and checking the sequence of the data (the file must contains
		// the data in a orderd way; matrixes are identified by a natural 
		// integer number sequence, starting from 0

		try{
		    
		    // read the file
		    BufferedReader fileMemoryMatrixes = null;
		    StringTokenizer t;
		    int matrixNumber, rows, cols;
		    boolean sensitive=false;
		    fileMemoryMatrixes = new 
			BufferedReader(new FileReader("unitData0/memoryMatrixes.txt"));
		    
		    // reading a control line (titles)
		    String line = " ", lineTest =
			"number(from_0_ordered)_rows_cols";

		    line=fileMemoryMatrixes.readLine();
		    if (! line.equals(lineTest))
			{
			    System.out.println(
				  "First line in unitData0/unitMemoryMatrixes.txt must contain:");
			    System.out.println(lineTest);
			    MyExit.exit(1);
			}
		    line=" ";
		    // starting with a void line; the actual
		    // reading is done in MyReader, dealing also
		    // with the missing data case
		    t = new StringTokenizer(line, " ");
		    
		    for(i = 0; i < totalMemoryMatrixNumber; i++)
			{
			    t=MyReader.check("unitData0/memoryMatrixes.txt", 
					     fileMemoryMatrixes, line, t);
			    if(! t.hasMoreTokens())
				lackingData("unitData0/memoryMatrixes.txt");
			    matrixNumber    = Integer.parseInt(t.nextToken());
			    if(matrixNumber < 0){
				System.out.println("You cannot use negative matrix codes here");
				MyExit.exit(1);
			    }
			    if(matrixNumber != i){
				System.out.println("Matrix numbers in unitData0/"+
						   "memoryMatrixes.txt must be a sequence"+
						   " of natural integer number "+
						   "starting from 0");
				MyExit.exit(1);
			    }
			    if(! t.hasMoreTokens())
				lackingData("unitData0/memoryMatrixes.txt");
			    rows            = Integer.parseInt(t.nextToken());
			    if(rows <= 0){
				System.out.println("Row numbers in unitData0/"+
						   "memoryMatrixes.txt cannot be <=0.");
				MyExit.exit(1);
			    }
			    if(! t.hasMoreTokens())
				lackingData("unitData0/memoryMatrixes.txt");
			    cols            = Integer.parseInt(t.nextToken());
			    if(cols <= 0){
				System.out.println("Column numbers in unitData/"+
						   "memoryMatrixes.txt cannot be <=0.");
				MyExit.exit(1);
			    }
			    
			    if(InformationRepository.getCountFromHibernatedMemoryMatrixList()==0)	
				aMatrix = new MemoryMatrix(
							   getZone(),i,sensitive,
							   (int) 1, rows,cols,
							   printMatrixes); 
			                                   // (int) 1 represents the number of order
			                                   // layers in jesframe; here always 1
			                                   // the boolean sensitive is related to
			                                   // layers and matrixes in jesframe; here
			                                   // it is always false
			    else 
				{
				    aMatrix=InformationRepository.removeFromHibernatedMemoryMatrixList();
				    aMatrix.init(getZone(),(int) -1, false,    // -1 conventional n. for 
				                                               // unit matrixes; sensitive is false
				       (int) 1, (int) 10, (int) 10,            // 1 is the unique layer; 10 and 10 r
				                                               // rows and cols
				       printMatrixes);
				}
			    matrixArray.atOffset$put(i,aMatrix);
			}

		    fileMemoryMatrixes.close();
		}
		catch (FileNotFoundException fnfe)
		    {
			System.out.println(fnfe);
			MyExit.exit(1);
		    } 
		catch (IOException e)
		    {
			System.out.println("IOException:" + e.toString());
			MyExit.exit(1);
		    }
	    } // end if(stratum==0)

	else  // stratum>0
	    {
		// copying memory matrix adresses
		for (int imm=0;imm<totalMemoryMatrixNumber;imm++)
		    matrixArray.atOffset$put(imm,StartESFrame.eSFrameObserverSwarm.
					     getMemoryMatrix(imm));

		if(StartESFrame.verbose)
		    {
			System.out.println("Copying memory matrixes in stratum "+stratum);
			for (int imm=0;imm<totalMemoryMatrixNumber;imm++)
			    System.out.println(((MemoryMatrix) getMemoryMatrix(imm)).getNumberOfRows()+" "+
					       ((MemoryMatrix) getMemoryMatrix(imm)).getNumberOfCols());
		    }
	    } // end else, i.e. stratum>0

	} // end if(totalMemoryMatrixNumber > 0)


	return this;
    }

    /**
     * Here is where the model schedule is built, the data structures
     * that define the simulation of time in the model. The core is an
     * actionGroup that has a list of actions. That's then put in a
     * Schedule.  */
    public Object buildActions ()
    {
	int i, ii;
	super.buildActions();

	modelActions2 = new ActionGroupImpl (getZone ());
	modelActions2generator = new ActionGroupImpl (getZone ());
	modelActions2distiller = new ActionGroupImpl (getZone ());
	modelActions2b = new ActionGroupImpl (getZone ());

	/* cleaning the unit activity space */
	modelActions2.createActionTo$message(unitActivitySpace,
			         SwarmUtils.getSelector(unitActivitySpace,"fillWithZero"));

	/* producing */
	modelActions2.createActionForEach$message(unitList,
			         SwarmUtils.getSelector("Unit","unitStep1"));
	/* a new order; this step is placed here, after the production step,
	 * to align the diffusion of the order forms (orders under execution 
	 * - next step - or new ones - this step -) */
	if(! useOrderDistiller)
	    modelActions2generator.createActionTo$message(orderGenerator,
			  SwarmUtils.getSelector(orderGenerator,
				 "createRandomOrderWithNSteps"));
	if(useOrderDistiller)
	    modelActions2distiller.createActionTo$message(orderDistiller,
			  SwarmUtils.getSelector(orderDistiller,
			  "distill"));
	modelActions2b.createActionForEach$message(unitList,
			  SwarmUtils.getSelector("Unit","unitStep2"));

	/* freeing the orders needing computations:
	 * these are orders generated or progagated both immediately above 
	 * and in previous generation and propagation steps */
	modelActions2b.createActionForEach$message(computationalAssemblerList,
		          SwarmUtils.getSelector("ComputationalAssembler",
				     "checkingComputationsAndFreeingOrders"));

	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"dropUnits"));
	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"generateANewUnit"));
	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"shuffleUnitList"));
	modelActions2b.createActionTo$message(this,
			  SwarmUtils.getSelector(this,"shuffleComputationalAssemblerList"));
	if(stratum==0 && stratumNumber>1)
	    modelActions2b.createActionTo$message(this,
		       SwarmUtils.getSelector(this,"shuffleModels"));

	// Then we create a schedule that executes the
	// modelActions. modelActions is an ActionGroup, by itself it
	// has no notion of time. In order to have it executed in
	// time, we create a Schedule that says to use the
	// modelActions ActionGroup at particular times
	modelSchedule = new ScheduleImpl (getZone (), 1);

	modelSchedule.at$createAction (0, modelActions2);

	if(! useOrderDistiller)
	    modelSchedule.at$createAction (0, modelActions2generator);
	else modelSchedule.at$createAction (0, modelActions2distiller);

	modelSchedule.at$createAction (0, modelActions2b);
	
	return this;
    }

    /**
     * Now set up the model's activation. swarmContext indicates where
     * we're being started in - typically, this model is run as a
     * subswarm of an observer swarm. */
    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves via the superclass
	// activateIn: method.  Just pass along the context: the
	// activity library does the right thing.
	super.activateIn (swarmContext);

	// Now activate our own schedule.
	modelSchedule.activateIn (this);

	// Finally, return our activity.
	return getActivity ();
  }
    
    /** unit generation */
    public Unit generateUnit(int i, int rep)
    {
	int xPos, yPos, xD=0, yD=0, count;
	Unit aUnit;
	
	xPos=Globals.env.uniformIntRand.getIntegerWithMin$withMax(
								  0, unitSpaceXSize-1);
	yPos=Globals.env.uniformIntRand.getIntegerWithMin$withMax(
								  0, unitSpaceYSize-1);
	
	// if the destination target is not empty, chose a direction ...
	if (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		xD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		yD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		
		while (xD==0 && yD==0)
		    {		
			xD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
			yD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
		    }
	    }
	// ... and move
	count=0;
	while (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		count++;
		// unitSpaceXSize+unitSpaceYSize is a raw test
		if (count >= unitSpaceXSize + unitSpaceYSize)return null;

		xPos += xD; xPos=CoordinateManager.check(unitSpaceXSize, xPos);	    
		yPos += yD; yPos=CoordinateManager.check(unitSpaceYSize, yPos);
	    }
	
	ComputationalAssembler aComputationalAssembler;
	if(InformationRepository.getCountFromHibernatedComputationalAssemblerList()==0)	
	    aComputationalAssembler = new ComputationalAssembler 
		(getZone(), orderDistiller, this);
	else 
	    {
		aComputationalAssembler=InformationRepository.
		    removeFromHibernatedComputationalAssemblerList();
		aComputationalAssembler.init(getZone(), orderDistiller, this);
	    }

	if(InformationRepository.getCountFromHibernatedMemoryMatrixList()==0)	
	    aMatrix = new MemoryMatrix(getZone(),(int) -1, false,    // -1 conventional n. for 
		 		                                     // unit matrixes; sensitive is false
				       (int) 1, (int) 10, (int) 10,  // 1 is the unique layer; 10 and 10 r
				                                     // rows and cols
				       printMatrixes); 
	else 
	    {
		aMatrix=InformationRepository.removeFromHibernatedMemoryMatrixList();
		aMatrix.init(getZone(),(int) -1, false,              // -1 conventional n. for 
				                                     // unit matrixes; sensitive is false
				       (int) 1, (int) 10, (int) 10,  // 1 is the unique layer; 10 and 10 r
				                                     // rows and cols
				       printMatrixes);
	    }
	aMatrix.increaseNumberOfUnitsUsingThisMemoryMatrix();
	// filling the matrix with 0
	for (int mr=0;mr<10;mr++) for (int mc=0; mc<10; mc++) aMatrix.setValue(0,mr,mc,0);

	if(InformationRepository.getCountFromHibernatedUnitList()==0)	
	    aUnit = new Unit  (getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       Globals.env.uniformDblRand.
			       getDoubleWithMin$withMax(
							minStartingVisibility, 
							maxStartingVisibility),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	else 
	    {
		aUnit=InformationRepository.removeFromHibernatedUnitList();
		aUnit.init(getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       Globals.env.uniformDblRand.
			       getDoubleWithMin$withMax(
							minStartingVisibility, 
							maxStartingVisibility),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	    }
	unitParameters.setParametersTo(i, rep, aUnit);
	aComputationalAssembler.setUnit(aUnit);

	
	// add the unit to the list
	unitList.addLast (aUnit);
	// placing the unit into the space
	unitSpace.putObject$atX$Y(aUnit, xPos, yPos);
	// add aComputationalAssembler to the list
	computationalAssemblerList.addLast(aComputationalAssembler);

	return aUnit;
	
    } // end generateUnit

    /** unit copying */
    public Unit copyUnit(Unit oldUnit, int rep)
    {
	int xPos, yPos, xD=0, yD=0, count;
	Unit aUnit;
	
	xPos=oldUnit.getXPos();
	yPos=oldUnit.getYPos();
	
	// if the destination target is not empty, chose a direction ...
	if (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		xD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		yD=Globals.env.uniformIntRand.getIntegerWithMin$withMax(-1, 1);
		
		while (xD==0 && yD==0)
		    {		
			xD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
			yD=Globals.env.uniformIntRand.
			    getIntegerWithMin$withMax(-1, 1);
		    }
	    }
	// ... and move
	count=0;
	while (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)
	    {
		count++;
		// unitSpaceXSize+unitSpaceYSize is a raw test
		if (count >= unitSpaceXSize + unitSpaceYSize)return null;

		xPos += xD; xPos=CoordinateManager.check(unitSpaceXSize, xPos);	    
		yPos += yD; yPos=CoordinateManager.check(unitSpaceYSize, yPos);
	    }
	
	ComputationalAssembler aComputationalAssembler;
	if(InformationRepository.getCountFromHibernatedComputationalAssemblerList()==0)	
	    aComputationalAssembler = new ComputationalAssembler 
		(getZone(), orderDistiller, this);
	else 
	    {
		aComputationalAssembler=InformationRepository.
		    removeFromHibernatedComputationalAssemblerList();
		aComputationalAssembler.init(getZone(), orderDistiller, this);
	    }

	if(InformationRepository.getCountFromHibernatedMemoryMatrixList()==0)	
	    aMatrix = new MemoryMatrix(getZone(),(int) -1, false,    // -1 conventional n. for 
				                                     // unit matrixes; sensitive is false
				       (int) 1, (int) 10, (int) 10,  // 1 is the unique layer; 10 and 10 r
				                                     // rows and cols
				       printMatrixes); 
	else 
	    {
		aMatrix=InformationRepository.removeFromHibernatedMemoryMatrixList();
		aMatrix.init(getZone(),(int) -1, false,              // -1 conventional n. for 
				                                     // unit matrixes; sensitive is false
				       (int) 1, (int) 10, (int) 10,  // 1 is the unique layer; 10 and 10 r
				                                     // rows and cols
				       printMatrixes);
	    }
	aMatrix.increaseNumberOfUnitsUsingThisMemoryMatrix();
	// filling the matrix with 0
	for (int mr=0;mr<10;mr++) for (int mc=0; mc<10; mc++) aMatrix.setValue(0,mr,mc,0);

	if(InformationRepository.getCountFromHibernatedUnitList()==0)	
	    aUnit = new Unit  (getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       Globals.env.uniformDblRand.
			       getDoubleWithMin$withMax(
							minStartingVisibility, 
							maxStartingVisibility),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	else
	    {
		aUnit=InformationRepository.removeFromHibernatedUnitList();
		aUnit.init(getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       Globals.env.uniformDblRand.
			       getDoubleWithMin$withMax(
							minStartingVisibility, 
							maxStartingVisibility),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	    }
	// instead of "unitParameters.setParametersTo(i, rep, aUnit);"
	int n=oldUnit.getUnitNumber()/100; n=n*100+rep;
	aUnit.setUnitNumber(n);
	aUnit.setProductionPhase(oldUnit.getProductionPhase()); 

	aComputationalAssembler.setUnit(aUnit);
	
	// add the unit to the list
	unitList.addLast (aUnit);
	// placing the unit into the space
	unitSpace.putObject$atX$Y(aUnit, xPos, yPos);
	// add aComputationalAssembler to the list
	computationalAssemblerList.addLast(aComputationalAssembler);
	
	return aUnit;

	} // end copyUnit 


    /** unit copying to another stratum */
    public Unit copyUnitToAnotherStratum(Unit oldUnit, int rep, int addendum,
					 int toStratum)
    {
	Unit aUnit;

	aUnit=((ESFrameModelSwarm) InformationRepository.getESFrameModelSwarm(toStratum)).
	    copyUnitInTheSamePlaceFromOtherStratum(oldUnit, rep, addendum);

	return aUnit;

	} // end copyUnitToAnotherStratum


    /** unit copying InTheSamePlaceFromOtherStratum */
    public Unit copyUnitInTheSamePlaceFromOtherStratum(Unit oldUnit, int rep, 
						       int addendum)
    {
	int xPos, yPos, xD=0, yD=0, count;
	Unit aUnit;
	
	xPos=oldUnit.getXPos();
	yPos=oldUnit.getYPos();
	
	// if the destination target is not empty, we do not create the new unit
	// without any advice
	if (unitSpace.getObjectAtX$Y(xPos,yPos)!=null)return null;
	
	ComputationalAssembler aComputationalAssembler;
	if(InformationRepository.getCountFromHibernatedComputationalAssemblerList()==0)	
	    aComputationalAssembler = new ComputationalAssembler 
		(getZone(), orderDistiller, this);
	else 
	    {
		aComputationalAssembler=InformationRepository.
		    removeFromHibernatedComputationalAssemblerList();
		aComputationalAssembler.init(getZone(), orderDistiller, this);
	    }

	aMatrix = oldUnit.getMemoryMatrix();
	aMatrix.increaseNumberOfUnitsUsingThisMemoryMatrix();

	if(InformationRepository.getCountFromHibernatedUnitList()==0)	
	    aUnit = new Unit  (getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       oldUnit.getVisibility(),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	else 
	    {
		aUnit=InformationRepository.removeFromHibernatedUnitList();
		aUnit.init(getZone(), this, unitActivitySpace, 
			       assigningTool, xPos, yPos, unitColor,
			       oldUnit.getVisibility(),
			       increasingVisibilityStep,
			       aComputationalAssembler, aMatrix, maxTicksAsUnsent);
	    }

	// instead of "unitParameters.setParametersTo(i, rep, aUnit);"
	int n=oldUnit.getUnitNumber(); // the same number
	aUnit.setUnitNumber(n);
	aUnit.setProductionPhase(oldUnit.getProductionPhase()+addendum); 

	aComputationalAssembler.setUnit(aUnit);
	
	// add the unit to the list
	unitList.addLast (aUnit);
	// placing the unit into the space
	unitSpace.putObject$atX$Y(aUnit, xPos, yPos);
	// add aComputationalAssembler to the list
	computationalAssemblerList.addLast(aComputationalAssembler);
	
	return aUnit;

	} // end copyUnitInTheSamePlaceFromOtherStratum 

    public Unit generateANewUnit(){
	int i;

	if(newUnitGenerationP >= 
		   Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		i=Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(1,proFormaUnitList.getCount());

		return generateUnit(i, 0); // no rep value, so the unit number will be i*100
	    }
	return null;
    }

    public Unit generateANewUnitFromAComputationalStep(){
	int i;

	i=Globals.env.uniformIntRand.
	    getIntegerWithMin$withMax(1,proFormaUnitList.getCount());
	
	return generateUnit(i, 0); // no rep value, so the unit number will be i*100
    }
    
    /** the method returns the list of the units */
    public Object getUnitList(){
	return unitList;
    }

    /** shuffling the unitList */
    public void shuffleUnitList(){
	myShuffler.shuffleWholeList(unitList);

	/*
	// to check the effect of the shuffling action
	  unitListIndex.setOffset(0);
	  for (int i=0;i<unitList.getCount();i++)
	    {
		System.out.print(((Unit)unitListIndex.get()).getUnitNumber()+" ");
		unitListIndex.next();
	    }
	System.out.println(" ");
	*/
    }

    /** shuffling the computationalAssemblerList */
    public void shuffleComputationalAssemblerList(){

	if(InformationRepository.getComputationalStepsInUse())
	    myShuffler.shuffleWholeList(computationalAssemblerList);
    }

    /** shuffling the models */
    public void shuffleModels()
    {
	InformationRepository.setModelSequence();
    }

    /** dropping units */
    public void dropUnits()
    {

	int i, nu, j, unsent=0, waiting=0, unsentValue=0, waitingValue=0;
	Unit aUnit;
	Order anOrder;

	nu=unitList.getCount();

	for (i=0;i<nu;i++)
	    {
		// the || is introduced to take in consideration a rare but possible
		// situation: the inactvity is at a limit value, but an order is just arrived
		aUnit=(Unit) unitList.removeFirst();

		if((! aUnit.getDroppable()) 
		   && (aUnit.getInactivity()<=maxInactivity || aUnit.getWaitingListLength()>0 
		                                            || maxInactivity==0)
                   && (aUnit.getUnsentProducts() <= maxUnsentProducts || maxUnsentProducts==0))
		    unitList.addLast(aUnit);
		else
		    {
			unitSpace.putObject$atX$Y(null, aUnit.getXPos(), aUnit.getYPos());
			
			unsentValue=0;
			unsent=aUnit.getUnsentProducts();
			for (j=1;j<=unsent;j++)
			    {
			    anOrder = (((Order)((ListImpl)aUnit.getDailyProductionList()).removeFirst()));
			    unsentValue += anOrder.getNumberOfDirectStepsDone();
			    anOrder.drop();
			    if(StartESFrame.verbose)
				System.out.println(anOrder.getOrderNumber()+" u  "+unsentValue);
			    }

			waitingValue=0;
			waiting=aUnit.getWaitingListLength();
			for (j=1;j<=waiting;j++)
			    {
			    anOrder = (((Order)((ListImpl)aUnit.getWaitingList()).removeFirst()));
			    waitingValue += anOrder.getNumberOfDirectStepsDone();
			    anOrder.drop();
			    if(StartESFrame.verbose)
				System.out.println(anOrder.getOrderNumber()+" w  "+waitingValue);
			    }

			addToGlobalProduction(-unsentValue-waitingValue);

			computationalAssemblerList.remove(aUnit.getComputationalAssembler());
			(aUnit.getComputationalAssembler()).drop();
			(aUnit.getMemoryMatrix()).drop();
			aUnit.drop();
		    }
	    }

	if(unitList.getCount()==0 && stopIfZeroUnits)
	    {
		System.out.println();
		System.out.println("Last unit dropped in stratum "+stratum+"; press Quit to exit.");
		StartESFrame.eSFrameObserverSwarm.
		    getControlPanel().setStateStopped();
	    }


    }


    /** record total production time and total recipe length */
    public void setProductionTimeAndRecipeLength(float pt, float rl) {
	totalProductionTime += pt;
	totalRecipeLength   += rl;
	if(StartESFrame.verbose)
	    System.out.println("totalProductionTime " + totalProductionTime +
			       " totalRecipeLength " + totalRecipeLength +
			       " ratio " + ((float)
					    totalProductionTime)/totalRecipeLength);
	return;
    }

    /** report the ratio totalProductionTime/totalRecipeLength */
    public float getTimeLengthRatio() {
	if(totalRecipeLength != 0)
	    return totalProductionTime/totalRecipeLength;
	else return (float) 1; // 1 is a good choice, being 0 totalRecipeLength
	                       // in the first simulation days and being 1
	                       // the actual ratio of the first
	                       // concluded orders
    }

    /** report the number of units */
    public float getNumberOfUnits(){
	return (float) unitList.getCount();
    }

    /** return the unitSpace */
    public Grid2dImpl getUnitSpace()
    {
	return unitSpace;
    }

    /** increasing potential production value */
    public void addToPotentialProduction(int d){
	 potentialProduction += d;
    }

    /** reporting potential production value */
    public float getPotentialProduction(){
	return potentialProduction;
    }

    /** increasing/decreasing global production value */
    public void addToGlobalProduction(int d){
	globalProduction += d;
    }

    /** reporting the global production value */
    public float getGlobalProduction(){
	return globalProduction;
    }

    /** increasing/decreasing the value of the concluded orders */
    public void addToConcludedOrders(float d){
	concludedOrders += d;
    }

    /** reporting the value of the concluded orders */
    public float getConcludedOrderSteps(){
	return concludedOrders;
    }
    /** return the unitActivitySpace address */
    public UnitActivitySpace getUnitActivitySpace()
    {
	return unitActivitySpace;
    }

    /** return the address of a memory matrix */
    public MemoryMatrix getMemoryMatrix(int matrixNumber){
	if(matrixNumber<0 || matrixNumber>=totalMemoryMatrixNumber)
	    {
		System.out.println("Error getting a matrix address.");
		MyExit.exit(1);
	    }
	return (MemoryMatrix) matrixArray.atOffset(matrixNumber);
    }

    /** return the number of memory matrixes */
    public int getTotalMemoryMatrixNumber(){
	return totalMemoryMatrixNumber;
	    }

    /** option about the matrix print capabulity */
    public void setMemoryMatrixPrintCapability(boolean pm){
	printMatrixes=pm;
    }

    /** conclude execution if data are lacking */
    public void lackingData(String s){
	System.out.println("Lacking data in " + s);
	MyExit.exit(1);
    }

    /** setting the stratum of this model */
    public void setStratum(int l, int ln)
    {
	stratum=l;
	stratumNumber=ln;
    }

    /** getting the stratum of this model */
    public int getStratum()
    {
	return stratum;
    }

    /** return if we want to create the numberOfUnitsGraph */
    public boolean getCreateNumberOfUnitsGraph()
    {
	return createNumberOfUnitsGraph;
    }

    /** return if we want to create the productionGraph */
    public boolean getCreateProductionGraph()
    {
	return createProductionGraph;
    }

    /** return if we want to create the ratioGraph */
    public boolean getCreateRatioGraph()
    {
	return createRatioGraph;
    }

    /** getting the maxInactivity value */
    public int getMaxInactivity()
    {
	return maxInactivity;
    }

    /** getting the maxUnsentProducts value */
    public int getMaxUnsentProducts()
    {
	return maxUnsentProducts;
    }

}
